#include <string.h>
#include <fftw.h>
#include <math.h>
#include "gaussian.h"

int cir_column (FFTW_COMPLEX *c, FFTW_COMPLEX *s, int i, int N1, int N2){
/* Suppose c is the first column of the 4N1*N2 \times 4N1*N2
   symmetrical block-circulant matrix with 2N1 block-circulant blocks
   of the size 2N2 \times 2N2 each.
   Put the i-th column of this matrix into s (already initialized).*/
   int k, block, pos;
   k=0;
   for (block = i / (2*N2); block >= 0; block--) {
      for (pos = i % (2*N2); pos >= 0; pos--) {
         c_re(s[k]) = c_re(c[2*N2*block + pos]);
         c_im(s[k]) = c_im(c[2*N2*block + pos]);
         k++;
      }
      for (pos = 2*N2 - 1; pos > i % (2*N2); pos--) {
         c_re(s[k]) = c_re(c[2*N2*block + pos]);
         c_im(s[k]) = c_im(c[2*N2*block + pos]);
         k++;
      }
   }
   for (block = 2*N1 - 1; block > i / (2*N2); block--) {
      for (pos = i % (2*N2); pos >= 0; pos--) {
         c_re(s[k]) = c_re(c[2*N2*block + pos]);
     c_im(s[k]) = c_im(c[2*N2*block + pos]);
     k++;
      }
      for (pos = 2*N2 - 1; pos > i % (2*N2); pos--) {
         c_re(s[k]) = c_re(c[2*N2*block + pos]);
         c_im(s[k]) = c_im(c[2*N2*block + pos]);
         k++;
      }
   }
   return 0;
}

double scal_prod (FFTW_COMPLEX *c, FFTW_COMPLEX *s, int i, int N1, int N2){
/* Suppose c is the first column of the 4N1*N2 \times 4N1*N2
symmetrical block-circulant matrix with 2N1 block-circulant blocks
of the size 2N2 \times 2N2 each.
Return the scalar product: the i-th column of C times s.
*/
   int k, block, pos;
   double result;
   result=0;
   k=0;
   for (block = i / (2*N2); block >= 0; block--) {
      for (pos = i % (2*N2); pos >= 0; pos--) {
         result += c_re(c[2*N2*block + pos]) * c_re(s[k]);
         k++;
      }
      for (pos = 2*N2 - 1; pos > i % (2*N2); pos--) {
         result += c_re(c[2*N2*block + pos]) * c_re(s[k]);
         k++;
      }
   }
   for (block = 2*N1 - 1; block > i / (2*N2); block--) {
      for (pos = i % (2*N2); pos >= 0; pos--) {
         result += c_re(c[2*N2*block + pos]) * c_re(s[k]);
         k++;
      }
      for (pos = 2*N2 - 1; pos > i % (2*N2); pos--) {
         result += c_re(c[2*N2*block + pos]) * c_re(s[k]);
         k++;
      }
   }
   return result;
}

double cov (int i, int j, int cols, double *theta, char covtype){
  /* returns entry i, j of the covariance
     matrix of the grid with cols columns.
     covtype='s' means spherical covariance
             'm' matern
             'r' rational quadratic
             'e' exponential
  */
  int x1, x2, y1, y2, nb, ncalc, ize;
  double kalpha, x;
  double l;
  double scaled_l;
  double *bk, ret;
  x1 = i % cols;
  y1=  i / cols;
  x2 = j % cols;
  y2=  j / cols;
  l = hypot(x2-x1, y2-y1);
  switch(covtype){
    case 'c': /*custom*/
       if (i==j) return 1; else return 0;
       break;
    case 'e':  /*exponential*/
               return pow(theta[0], pow(l, theta[1]));
               break;
    case 'r': /*rational quadratic */
              if (l == 0) return 1;
              scaled_l = l / theta[0];
              return pow(1 + scaled_l*scaled_l, -theta[1]);
              break;
    case 's': /* spherical */
              if (l > *theta) return 0;
              else {
                scaled_l = l / (*theta);
                return 1 - 1.5 * scaled_l + 0.5*scaled_l*scaled_l*scaled_l;
              }
              break;
    case 'm': /* Matern  */
             if (l == 0) return 1;
             nb = (int) theta[1];
             kalpha = theta[1] - nb;
             bk = (double *) malloc(++nb *sizeof(double));
             if (bk==NULL) return 10;
             x = l/theta[0];
             ize=2;
             rkbesl_(&x, &kalpha, &nb, &ize, bk, &ncalc);
             if (ncalc != nb) {
                /*printf("Error: could not compute modified bessel function K\n");*/
                free(bk);
                return 10;
             }
             ret = 2*exp(theta[1]*log(x/2)+log(bk[nb-1])-x-lgamma(theta[1]));
             if (ret < 0 || ret > 1) ret=10;
             free(bk);
             return ret;
             break;
  }
}



FFTW_COMPLEX *generate_w_given_x (ghandle handle,
                                  int gibbs_iterations,
                                  double *enc_a,
                                  double *enc_b,
                                  double *theta,
                                  char covtype){
  /*   Generate Gaussian vector W given enc_a < W < enc_b,
       where W encompasses m1*m2-vector Z.
  gibbs_iterations: number of Gibbs iterations to perform;
  *enc_a, *enc_b: conditions.
  *theta, covtype: covariance parameters and family.
  Returns pointer to the resulting field (handle -> w) */
  int i, j, l, m1, m2;
  int posdef;
  double var;
  m1=handle->n1;
  m2=handle->n2;
  fill_in_encompassing_matrix (handle, theta, covtype);
  fftwnd(handle->plan, 1, (FFTW_COMPLEX *) handle->k, 1, 0,
                          (FFTW_COMPLEX *) NULL, 1, 0);
  posdef=1;
  for (i=0; i<4*m1*m2; i++){
    if (c_re(handle->k[i])<=0){
      posdef=0;
      break;
    }
    c_re(handle->k[i])=1/c_re(handle->k[i]);
    c_im(handle->k[i])=0;
  }
  if (posdef==0) {
    printf ("Encompassing matrix was not positive-definite.\n");
    return NULL;
  }
  else {  /* start gibbs */
     /*put initial configuration into y */
     for (i=0; i < 4*m1*m2; i ++) {
        c_re(handle->w[i])=normal(0, 1);
        c_im(handle->w[i])=0;
     }
     /*put the first column of C inverse into k*/
     fftwnd(handle->planb, 1, (FFTW_COMPLEX *) handle->k, 1, 0,
                              (FFTW_COMPLEX *) NULL, 1, 0);
     for (i=0; i < 4*m1*m2; i++) {
       c_re(handle->k[i]) /= 4*m1*m2;
     }
     /* do gibbs iterations */
     var=1/c_re(handle->k[0]);
     for (l=0; l<gibbs_iterations; l++) {
      for (i=0; i < 2*m1; i++){
        for (j=0; j < 2*m2; j++) {  /* update component w[i,j]*/
          c_re(handle->w[i*2*m2 + j])=0;
          c_re(handle->w[i*2*m2 + j])=
            trunc_normal(-scal_prod (handle->k, handle->w,
                                     i*2*m2 + j, m1, m2)/c_re(handle->k[0]),
                          var, enc_a[i*2*m2 + j], enc_b[i*2*m2 + j]);
        }
      }
     }
     return handle->w;
  }
}


FFTW_COMPLEX *fill_in_encompassing_matrix (
            ghandle handle, double *theta, char covtype){
  /* Fills the encompassing cov matrix for the field
     with cov. parameters *theta, covtype.
     Returns pointer to the resulting matrix (handle -> k) */
  int l, i, m1, m2;
  m1=handle->n1;
  m2=handle->n2;
  for (l=0; l<m1; l++)   {
     for (i=0; i<m2; i++) {
       c_re(handle->k[l*2*m2 + i]) = cov(0, l*m2 +i, m2, theta, covtype);
       if ((c_re(handle->k[l*2*m2 + i])==10) && (covtype=='m')) {
         handle->status=-3;
         return NULL;
       }
       c_im(handle->k[l*2*m2 + i]) = 0;
     }
     c_re(handle->k[l*2*m2 + m2]) = 0;
     c_im(handle->k[l*2*m2 + m2]) = 0;
     for (i=1; i<m2; i++) {
         c_re(handle->k[l*2*m2 + 2*m2 - i]) = cov(i, l*m2, m2, theta, covtype);
         if ((c_re(handle->k[l*2*m2 + i])==10) && (covtype=='m')) {
           handle->status=-3;
           return NULL;
         }
         c_im(handle->k[l*2*m2 + 2*m2 - i]) = 0;
     }
  }
  for (i=2*m2*m1; i < 2*m2*m1 + 2*m2; i++) {
        c_re(handle->k[i]) = 0;
        c_im(handle->k[i]) = 0;
  }
  for (l=m1-1; l>0; l--)   {
     for (i=0; i<m2; i++) {
       c_re(handle->k[4*m1*m2 - l*2*m2 + i]) = cov(i, l*m2, m2, theta, covtype);
       if ((c_re(handle->k[l*2*m2 + i])==10) && (covtype=='m')) {
         handle->status=-3;
         return NULL;
       }
       c_im(handle->k[4*m1*m2 - l*2*m2 + i]) = 0;
     }
     c_re(handle->k[4*m1*m2 - l*2*m2 + m2]) = 0;
     c_im(handle->k[4*m1*m2 - l*2*m2 + m2]) = 0;
     for (i=1; i<m2; i++) {
         c_re(handle->k[4*m1*m2 - l*2*m2 + 2*m2 - i]) = cov(0, l*m2+i, m2, theta, covtype);
         if ((c_re(handle->k[l*2*m2 + i])==10) && (covtype=='m')) {
           handle->status=-3;
           return NULL;
         }
         c_im(handle->k[4*m1*m2 - l*2*m2 + 2*m2- i]) = 0;
     }
  }
  return handle->k;
}

FFTW_COMPLEX *generate_w (ghandle handle){
  /*
    Generate two independent multivariate normal vectors with mean zero
    and block circulant covariance matrix with circulant blocks given
    by the first column handle->k.
    If k is not posdef returns -1;
    Otherwise puts two realizations of the vector into Re k and Im k
    Returns pointer to the result (handle -> k)
  */
  int i, m1, m2;
  double variance;
  double norm;
  m1=2*handle->n1;
  m2=2*handle->n2;
  fftwnd(handle->plan, 1, (FFTW_COMPLEX *) handle->k, 1, 0,
                          (FFTW_COMPLEX *) NULL, 1, 0);
  norm=sqrt(m1*m2);
  for (i=0; i < m1*m2; i++){
       variance = c_re(handle->k[i]);
       if (variance<=0) return NULL;
       c_re(handle->k[i]) = normal(0, variance);
       c_im(handle->k[i]) = normal(0, variance);
  }
  fftwnd(handle->plan, 1, (FFTW_COMPLEX *) handle->k, 1, 0,
                          (FFTW_COMPLEX *) NULL, 1, 0);
  for (i=0; i<m1*m2; i++) {
       c_re(handle->k[i]) /= norm;
       c_im(handle->k[i]) /= norm;
  }
  return handle->k;
}



int EM (double *a,
        double *b,
        int n1, int n2,
        int gibbs_iterations,
        int em_iterations,
        char covtype,
        double *theta,
        int ntheta,
        int usehooke,
        char *log_fname){
/* Estimate covariance parameter from the quantized data.
   Gaussian Z is not observed; however, it is known that a < Z < b.
   em_iterations - number of EM steps to perform.
   gibbs_iterations - number of Gibbs iterations to perform at each step.
*/
  int i, d, n, r;
  FILE *log;
  ghandle handle;
  double *theta_out;
  double *enc_a, *enc_b;

  enc_a=(double *) malloc(4*n1*n2*sizeof(double));
  enc_b=(double *) malloc(4*n1*n2*sizeof(double));
  for (i=0; i<4*n1*n2; i ++) {
         enc_a[i]=-1000;
         enc_b[i]=1000;
  }
  d=0;
  for (n=0; n<n1; n++)
       for (r=0; r<n2; r++){
         enc_a[2*n*n2+r] = a[d];
         enc_b[2*n*n2+r] = b[d++];
       }
  if (!(handle = create_ghandle(n1, n2, FFTW_MEASURE))) return -1;
  likelihood_params = (struct likelihood_params_struct *)
                          malloc(sizeof(struct likelihood_params_struct));
  likelihood_params->handle = handle;
  likelihood_params->covtype=covtype;
  log=fopen(log_fname, "w");
  fclose(log);
  theta_out = (double *) malloc (ntheta * sizeof(double));
  /* do EM iterations */
  for (i=0; i < em_iterations; i++) {
        printf ("%d. ",i);
        for (n=0; n<ntheta; n++)
           printf("theta[%d]=%f ", n, theta[n]);
        printf("\n");
        log=fopen(log_fname, "a");
        for (n=0; n<ntheta; n++)
           fprintf(log, "%f ",theta[n]);
        fprintf(log, "\n");
        fclose(log);
        /* w = sample with theta; theta =  arg max L(w|theta) */
        if (generate_w_given_x(handle, gibbs_iterations,
                               enc_a, enc_b, theta, covtype)) {
           fftwnd(handle->plan, 1, (FFTW_COMPLEX *) handle->w, 1, 0,
                                   (FFTW_COMPLEX *) NULL, 1, 0);
           if (usehooke) {
             hooke(minus_likelihood,
                   ntheta,
                   theta,
                   theta_out,
                   0.5,
                   1e-6,
                   5000);
             for (n=0; n<ntheta; n++)
                       theta[n]=theta_out[n];
           }
           else theta[0]=fminbr(1, 40, minus_likelihood_1param, 1e-6);
        }
        else break;
  }
  free(enc_a);
  free(enc_b);
  free (likelihood_params);
  destroy_ghandle(handle);
  return 0;
}

double minus_likelihood (double *theta, int n){
/* computes the likelihood of the observed transformed data.
   with covariance parameter theta.
   other parameters are taken from the global
      struct likelihood_params *likelihood_params_struct;
   For use with hooke.
*/
  double logdetK;
  double qform;
  int i, n1, n2;
  FFTW_COMPLEX *k, *w;
  n1=likelihood_params->handle->n1;
  n2=likelihood_params->handle->n2;
  k = likelihood_params->handle->k;
  w = likelihood_params->handle->w;
  fill_in_encompassing_matrix (likelihood_params->handle,
                               theta,
                               likelihood_params->covtype);
  fftwnd(likelihood_params->handle->plan, 1,
         (FFTW_COMPLEX *) k, 1, 0,
         (FFTW_COMPLEX *) NULL, 1, 0);
  logdetK=0;
  for (i=0; i<4*n1*n2; i++) {
     if (c_re(k[i])<=0) return 99e99;
     else logdetK=logdetK+log(c_re(k[i]));
  }
  qform=0;
  for (i=0; i<4*n1*n2; i++)
    qform=qform+(c_re(w[i])*
                 c_re(w[i]) +
                 c_im(w[i])*
                 c_im(w[i])) / c_re(k[i]);
  qform /= (4*n1*n2);
  return logdetK+qform;
}


double minus_likelihood_1param (double x){
  return minus_likelihood (&x,  1);
}

FFTW_COMPLEX *generate_z(ghandle handle,
                         double *theta,
                         char covtype){
  /* generate two samples from zero-mean gaussian random field
     with covariance function cov()
     Upon exit,
     Re z and Im z will contain independent realizations.
     returns poiinter to z (handle->w)*/
  int m1, m2;
  m1=handle->n1;
  m2=handle->n2;
  if (!fill_in_encompassing_matrix (handle, theta, covtype)) return NULL;
  if (generate_w (handle)) return z_from_w(handle);
  else {
   handle -> status=-1;
   return NULL;
  }
}

FFTW_COMPLEX *z_from_w (ghandle handle) {
/* Recovers field z (handle->w) from the encompassing field w (handle->k).
   Returns pointer to the result (handle->w)*/
   int l, i, j, m1, m2;
   m1=handle->n1;
   m2=handle->n2;
   l=0;
   for (i=0; i<m1; i++)
      for (j=0; j<m2; j++){
        c_re(handle->w[l])= c_re(handle->k[2*i*m2 + j]);
        c_im(handle->w[l++])= c_im(handle->k[2*i*m2 + j]);
      }
    return handle->w;
}

static int complexcompare(const void *i, const void *j)  {
   if (c_re(*(FFTW_COMPLEX *)i) > c_re(*(FFTW_COMPLEX *)j)) return 1;
   if (c_re(*(FFTW_COMPLEX *)i) < c_re(*(FFTW_COMPLEX *)j)) return -1;
   return 0;
}

int write_field (int n1,        /*for generating*/
                 int n2,        /*for generating*/
                 int wn1,        /*output window size*/
                 int wn2,        /*output window size*/
                 double *theta,  /*parameters*/
                 char covtype,
                 char *fname_gaussian_ppm,
                 char *fname_gaussian_num,
                 char *fname_clipped_ppm,
                 char *fname_clipped_num,
                 char format,
                 int nlevels,
                 int *r, int *g, int *b,
                 double *c,
                 int thresholds){
   /*
   Write a wn1 x wn2 field to file in various formats.
   m1, m2 - gird size used for generating to overcome negative
       embedding;
   theta, covtype - define covariance;
   fname_gaussian_ppm - if non-null, Gaussian field will be
                                                 written in PPM.
   fname_gaussian_num - if non-null, Gaussian field will be written in
                                                 numeric format.
   fname_clipped_ppm  - if non-null, clipped field will be
                                                 written in PPM.
   fname_clipped_num  - if non-null, clipped field will be written in
                                                  numeric format.
   format - specifies numeric file format, '3' or 'Z'.
   nlevels - number of thresholds for clipping; e.g. 1 for binary image.
   r[nlevels+1], g[nlevels+1], b[nlevels+1] -
             RGB levels for coloring clipped image;
   double c[nlevels] -thresholds for clipping. */
 ghandle handle;
 FFTW_COMPLEX *window;
 double sum;
 int i, j, ans;
 char comment_g[256], comment_c[256], buf[256];
 handle=create_ghandle(n1, n2, FFTW_ESTIMATE);
 if (!handle) return -2; /* out of memory */
 generate_z(handle, theta, covtype);
 if (!(ans=handle->status)) {
          if ((wn1 != n1) || (wn2 != n2)) {
            for (i=0; i<wn1; i++)
                for(j=0; j<wn2; j++) {
                   c_re(handle->k[i*wn2 + j]) = c_re(handle->w[i*n2+j]);
                   c_im(handle->k[i*wn2 + j]) = c_im(handle->w[i*n2+j]);
                }
            memcpy(handle->w, handle->k, wn1*wn2*sizeof(FFTW_COMPLEX));
          } else memcpy(handle->k, handle->w, wn1*wn2*sizeof(FFTW_COMPLEX));
          window=handle->w;
          if ((fname_clipped_num || fname_clipped_ppm) && !thresholds){
             qsort((char *) handle->k, wn1*wn2, sizeof(FFTW_COMPLEX), complexcompare);
             sum=0;
             for (i=0; i < nlevels; i++){
               sum += c[i];
               c[i]=c_re(handle->k[(int)rint(wn1*wn2*sum)+1]);
             }
          }
          switch (covtype){
            case 'r': sprintf(comment_g, "# Gaussian %dx%d r.f. \
with Rational Quadratic (%f, %f) correlation and mean 0.",
wn1, wn2, theta[0], theta[1]);
            break;
            case 's':sprintf(comment_g, "# Gaussian %dx%d r.f. \
with Spherical (%f) correlation and mean 0.",wn1, wn2, theta[0]);
            break;
            case 'e':sprintf(comment_g, "# Gaussian %dx%d r.f. \
with Exponential (%f, %f) correlation and mean 0.",
wn1, wn2, theta[0], theta[1]);
            break;
            case 'c':sprintf(comment_g, "# Gaussian %dx%d r.f. \
with Custom (%f, %f) correlation and mean 0.",
wn1, wn2, theta[0], theta[1]);
            break;
            case 'm':sprintf(comment_g, "# Gaussian %dx%d r.f. \
with Matern (%f, %f) correlation and mean 0.",
          wn1, wn2, theta[0], theta[1]);
          }
          sprintf(comment_c, "%s\n# Clipped at level(s) (", comment_g);
          for (i=0; i<nlevels-1; i++) {
            sprintf(buf, "%f, ", c[i]);
            strcat(comment_c, buf);
          }
          sprintf(buf,"%f)", c[nlevels-1]);
          strcat(comment_c, buf);
          if (fname_gaussian_ppm)
               write_ppm(window, 0, fname_gaussian_ppm, wn1, wn2);
          if (fname_clipped_ppm)
              write_clipped_ppm (window, 0, nlevels, c, r, g, b,
                            fname_clipped_ppm, wn1, wn2);
          if (fname_gaussian_num)
             write_numeric (window, 0, format,
                         fname_gaussian_num, wn1, wn2,comment_g);
          if (fname_clipped_num)
            write_clipped_numeric (window, 0, format,
                nlevels, c, fname_clipped_num, wn1, wn2, comment_c);
 }
 destroy_ghandle(handle);
 return ans;
}


ghandle create_ghandle(int n1, int n2, int method){
/* Creates a handle for n1 x n2 fields.
   method=FFTW_MEASURE: takes longer to produce faster handle
   method=FFTW_ESTIMATE: takes shorter to produce slower handle
*/
   ghandle ptr;
   if (!(ptr=malloc(sizeof(struct ghandle_struct)))) return NULL;
   ptr->n1=n1;
   ptr->n2=n2;
   if (!(ptr->plan=fftw2d_create_plan(2*n1, 2*n2,
     FFTW_FORWARD, method | FFTW_IN_PLACE))) return NULL;
   if (!(ptr->planb=fftw2d_create_plan(2*n1, 2*n2,
     FFTW_BACKWARD, method | FFTW_IN_PLACE))) return NULL;
   if (!(ptr->w  = (FFTW_COMPLEX *) malloc (4*n1*n2*sizeof(FFTW_COMPLEX))))
     return NULL;
   if (!(ptr->k  = (FFTW_COMPLEX *) malloc (4*n1*n2*sizeof(FFTW_COMPLEX))))
     return NULL;
   ptr->status=0;
   return ptr;
}

int destroy_ghandle(ghandle handle){
  if (handle) {
    free(handle->w);
    free(handle->k);
    fftwnd_destroy_plan(handle->plan);
    fftwnd_destroy_plan(handle->planb);
    free(handle);
  }
  return 0;
}
