#include "fftw.h"


/* ----------------------------------------------------------- */
/*          UNIVARIATE RANDOM VARIABLES     rv.c               */
/* ----------------------------------------------------------- */

/* generate unif [rmin,rmax] */
double runif (double rmin, double rmax);

/* generate a discrete RV taking n integer values
   v[] with probabilities p[] */
int discrv (int v[], double p[], int n);

/* generate standard normal */
double n01();

/* generate standard normal(mu, sigma2) */
double normal (double mu, double sigma2);

/* Phi, standard normal cdf */
double poz(double z);

/* Phi inverse, inverse of standard normal cdf      */
/* returns z such that fabs (poz(p) - z) <= .000001 */
double critz (double p);

/* generate Z ~ N(mu, sigma2) such that a<Z<b */
double trunc_normal (double mu, double sigma2, double a, double b);


#define Phi      poz
#define PhiInv   critz

/* ----------------------------------------------------------- */
/*                 FILE FORMATS                formats.c       */
/* ----------------------------------------------------------- */

/* Write image stored in inpic to a color ppm file, as grey levels.
   which=0 means use real part of inpic
   which=1 means use imaginary part of inpic
   image size is rows \times cols. */
int write_ppm(FFTW_COMPLEX *inpic,
              int which,
              char *fname,
              int rows,
              int cols);


int write_numeric (FFTW_COMPLEX *inpic,
                   int which,
                   char format,
                   char *fname,
                   int rows,
                   int cols,
                   char *comment);


int read_numeric (FFTW_COMPLEX *inpic,
                  int which,
                  char format,
                  char *fname,
                  int rows,
                  int cols);

/* inpic contains continuous field. Clip it at thresholds c
   and write resulting discrete field as color color ppm file,
   using color   (r[i], g[i], b[i]) iff inpic[i] is less than c[i].
   which=0 means use real part of inpic
   which=1 means use imaginary part of inpic
   image size is rows \times cols.
   e.g. for a three color image use
            nlevels=2
            c=(-0.43, 0.43)
            r=(153, 255, 51);
            g=(0,255,153);
            b=(0,0,255);
  */
int write_clipped_ppm (FFTW_COMPLEX *inpic,
                       int which,
                       int nlevels,
                       double *c,
                       int *r,
                       int *g,
                       int *b,
                       char *fname,
                       int rows,
                       int cols);

/* writes out clipped re(inpic) for which=0 or
              clipped im(inpic) for which=1;
The number of clipping thresholds is nlevels.
E.g. nlevels = 1 for binary data.
c[nlevels] holds the thresholds.
If format='3', the output format is
 x y z
 x y z
 x y z;
If format='Z', the output is Z values only.*/

int write_clipped_numeric (FFTW_COMPLEX *inpic,
                           int which,
                           char format,
                           int nlevels,
                           double *c,
                           char *fname,
                           int rows,
                           int cols,
                           char *comment);
/* writes out clipped re(inpic) for which=0 or
              clipped im(inpic) for which=1;
The number of clipping thresholds is nlevels.
E.g. nlevels = 1 for binary data.
c[nlevels] holds the thresholds.
If format='3', the output format is
 x y z
 x y z
 x y z;
If format='Z', the output is Z values only.*/

int get_constraints_ppm (char *fname,
                         int *rows,
                         int *cols,
                         int nlevels,
                         double *c,
                         int *r,
                         double **a,
                         double **b);
  /* Get image dimensions rows x cols and constraints (observed data)
     from ppm file *fname and levels. The nlevels, levels c, and red signals r of each level
     are given. Allocate memory and populate a, b so that a .< z .< b.
     e.g. c=(-0.43, 0.43) nlevels=2, three-color picture
          r=(153, 255, 51); but 51 here is not important and will
          not be used.
  */
int get_constraints_num (char *fname,
                         char file_format,
                         int rows,
                         int cols,
                         int nlevels,
                         double *c,
                         int *codes,
                         double **a,
                         double **b);
  /* For n1xn2 image,  get constraints (observed data)
     from numeric file *fname and levels.
     The nlevels, levels c, and codes of each level
     are given. Allocate memory and populate a, b so that a .< z .< b.
     e.g. for three color image, c=(-0.43, 0.43), nlevels=2,
          codes=(0, 1, 2); but 2 here is not important and will
          not be used.
  */
int getint_ini (char *fname, char *lhs);
/*
 Get int value from the string in file named fname of the form
 lhs=2 or lhs:2
*/

double getdouble_ini (char *fname, char *lhs);
/*
 Get double value from the string in file named fname of the form
 lhs=2 or lhs:2
*/

char *getstring_ini (char *fname, char *lhs);
/*
 Get string from the string in file named fname of the form
 lhs=2 or lhs:2
*/


/* ----------------------------------------------------------- */
/*                 GAUSSIAN RANDOM FIELDS          fields.c    */
/* ----------------------------------------------------------- */


#define MAX_THETA 10

struct ghandle_struct{
     FFTW_COMPLEX *w;
     FFTW_COMPLEX *k;
     int n1;
     int n2;
     fftwnd_plan plan;
     fftwnd_plan planb;
     int status;
};

typedef struct ghandle_struct *ghandle;

struct likelihood_params_struct{
     char covtype;
     ghandle handle;
};

struct likelihood_params_struct *likelihood_params;

int rkbesl_(double *x, double *alpha, int *nb,
               int *ize, double *bk, int *ncalc);

double fminbr(double a, double b,
              double (*f)(double x),
              double tol);

int hooke(double (*f)(double *x, int n),
          int nvars,
          double *startpt, double *endpt,
          double rho, double epsilon,
          int itermax);

double cov (int i, int j, int cols, double *theta, char covtype);
  /* returns entry i, j of the covariance
     matrix of the grid with cols columns.
     covtype='s' means spherical covariance
             'm' matern
             'r' rational quadratic
             'e' exponential
  */

FFTW_COMPLEX *generate_w_given_x (ghandle handle,
                                  int gibbs_iterations,
                                  double *enc_a,
                                  double *enc_b,
                                  double *theta,
                                  char covtype);
  /*   Generate Gaussian vector W given enc_a < W < enc_b,
       where W encompasses m1*m2-vector Z.
  gibbs_iterations: number of Gibbs iterations to perform;
  *enc_a, *enc_b: conditions.
  *theta, covtype: covariance parameters and family.
  Returns pointer to the resulting field (handle -> w) */

FFTW_COMPLEX *fill_in_encompassing_matrix (
       ghandle handle, double *theta, char covtype);
  /* Fills the encompassing cov matrix for the field
     with cov. parameters *theta, covtype.
     Returns pointer to the resulting matrix (handle -> k) */

FFTW_COMPLEX *generate_w (ghandle handle);
  /*
    Generate two independent multivariate normal vectors with mean zero
    and block circulant covariance matrix with circulant blocks given
    by the first column handle->k.
    If k is not posdef returns -1;
    Otherwise puts two realizations of the vector into Re k and Im k
    Returns pointer to the result (handle -> k)
  */

int EM (double *a,
        double *b,
        int n1,
        int n2,
        int gibbs_iterations,
        int em_iterations,
        char covtype,
        double *theta,
        int ntheta,
        int usehooke,
        char *log_fname);
/* Estimate covariance parameter from the quantized data.
   Gaussian Z is not observed; however, it is known that a < Z < b.
   em_iterations - number of EM steps to perform.
   gibbs_iterations - number of Gibbs iterations to perform at each step.
*/

double minus_likelihood (double *theta, int n);
/* computes the likelihood of the observed transformed data.
   with covariance parameter theta.
   other parameters are taken from the global
      struct likelihood_params *likelihood_params_struct;
   For use with hooke.
*/

double minus_likelihood_1param (double x);

FFTW_COMPLEX *generate_z(ghandle handle,
                         double *theta,
                         char covtype);
  /* generate two samples from zero-mean gaussian random field
     with covariance function cov()
     Upon exit,
     Re z and Im z will contain independent realizations.
     Returns pointer to z (handle->w)*/

FFTW_COMPLEX *z_from_w (ghandle handle);
/* Recovers field z (handle->w) from the encompassing field w (handle->k).
   Returns pointer to the result (handle->w)*/

int write_field (int n1,        /*for generating*/
                 int n2,        /*for generating*/
                 int wn1,        /*output window size*/
                 int wn2,        /*output window size*/
                 double *theta, /*parameters*/
                 char covtype,
                 char *fname_gaussian_ppm,
                 char *fname_gaussian_num,
                 char *fname_clipped_ppm,
                 char *fname_clipped_num,
                 char format,
                 int nlevels,
                 int *r, int *g, int *b,
                 double *c,
                 int thresholds);
   /*
   Write a wn1 x wn2 field to file in various formats.
   m1, m2 - gird size used for generating to overcome negative
       embedding;
   theta, covtype - define covariance;
   fname_gaussian_ppm - if non-null, Gaussian field will be
                                                 written in PPM.
   fname_gaussian_num - if non-null, Gaussian field will be written in
                                                 numeric format.
   fname_clipped_ppm  - if non-null, clipped field will be
                                                 written in PPM.
   fname_clipped_num  - if non-null, clipped field will be written in
                                                  numeric format.
   format - specifies numeric file format, '3' or 'Z'.
   nlevels - number of thresholds for clipping; e.g. 1 for binary image.
   r[nlevels+1], g[nlevels+1], b[nlevels+1] -
             RGB levels for coloring clipped image;
   double c[nlevels] - if (thresholds) : thresholds for clipping
                       otherwise : fractions of the corresponding (r,g,b) color in the image
   */

ghandle create_ghandle(int n1, int n2, int method);
/* Creates a handle for n1 x n2 fields.
   method=FFTW_MEASURE: takes longer to produce faster handle
   method=FFTW_ESTIMATE: takes shorter to produce slower handle
*/

int destroy_ghandle(ghandle handle);
